/**

Haven't formally verified this grammar but it should be LL(2)

The root's intermediate state has StartTag/EndTag because it's easier to first parse them as independant nodes
than to parse a StartTag and find the matching EndTag since we can only lookahead by 1 token

Trying to lookahead by 4 tokens after each advancement to determine the end of the sub-root will greatly affect performance
    1 "["
    2 "/"
    3 "LABEL"
    4 "]"

---

Root <- (Text | Linebreak | Tag)*

Text <-
| {XSS Characters}.
| STR.

Linebreak <-
| LINEBREAK.

Tag      <- StartTag Root EndTag
StartTag <- L_BRACKET           Text Attr* R_BRACKET
EndTag   <- L_BRACKET BACKSLASH Text       R_BRACKET

Attr <-
| STR EQUALS STR
| EQUALS STR
| STR

*/
export declare const enum AstNodeType {
    RootNode = 0,
    TextNode = 1,
    LinebreakNode = 2,
    TagNode = 3,
    StartTagNode = 4,
    EndTagNode = 5,
    AttrNode = 6
}
export declare function nodeTypeToString(nodeType: AstNodeType): string;
export declare abstract class AstNodeBase {
    readonly abstract nodeType: AstNodeType;
    readonly children: Array<AstNode>;
    constructor(children?: Array<AstNode>);
    addChild(node: AstNode): void;
    isValid(): boolean;
    toShortString(): string;
    toString(depth?: number): string;
}
export declare class RootNode extends AstNodeBase {
    readonly nodeType = AstNodeType.RootNode;
    isValid(): boolean;
}
export declare class TextNode extends AstNodeBase {
    readonly nodeType = AstNodeType.TextNode;
    readonly str: string;
    constructor(str: string);
    isValid(): boolean;
    toShortString(): string;
}
export declare class LinebreakNode extends AstNodeBase {
    readonly nodeType = AstNodeType.LinebreakNode;
    toShortString(): string;
}
export declare class StartTagNode extends AstNodeBase {
    readonly nodeType = AstNodeType.StartTagNode;
    readonly tagName: string;
    readonly ogTag: string;
    constructor(tagName: string, ogTag: string, attrNodes?: Array<AttrNode>);
    isValid(): boolean;
    toShortString(): string;
}
export declare class EndTagNode extends AstNodeBase {
    readonly nodeType = AstNodeType.EndTagNode;
    readonly tagName: string;
    readonly ogTag: string;
    constructor(tagName: string, ogTag: string);
    isValid(): boolean;
    toShortString(): string;
}
export declare class TagNode extends AstNodeBase {
    readonly nodeType = AstNodeType.TagNode;
    private readonly _startTag;
    private readonly _endTag?;
    constructor(startTag: StartTagNode, endTag?: EndTagNode | LinebreakNode);
    get tagName(): string;
    get attributes(): Array<AttrNode>;
    getAttributesByName(): Record<string, string>;
    get ogStartTag(): string;
    get ogEndTag(): string;
    isValid(): boolean;
    toString(depth?: number): string;
    /**
     * Gets the text of the immediate descendant of the current TagNode
     *
     * For example, would return "https://en.wikipedia.org/wiki/Markdown" for the following:
     * ```bbcode
     * [url]https://en.wikipedia.org/wiki/Markdown[/url]
     * ```
     */
    getTagImmediateText(this: TagNode): string | undefined;
    /**
     * Gets the value of the immediate attribute of the current TagNode
     *
     * For example, would return "https://en.wikipedia.org/wiki/Markdown" for the following:
     * ```bbcode
     * [url=https://en.wikipedia.org/wiki/Markdown] Wikipedia's page on Markdown [/url]
     * ```
     */
    getTagImmediateAttrVal(this: TagNode): string | undefined;
}
export declare class AttrNode extends AstNodeBase {
    readonly nodeType = AstNodeType.AttrNode;
    static readonly DEFAULT_KEY = "default";
    get key(): string;
    get val(): string;
    isValid(): boolean;
    toShortString(): string;
}
export type AstNode = RootNode | TextNode | LinebreakNode | TagNode | StartTagNode | EndTagNode | AttrNode;
//# sourceMappingURL=AstNode.d.ts.map